#
import torch
from torch import nn
import threading as th
from ... import Base
import numpy as np
class Fcs(Base):
    def init(self, fcs):
        self.fcs = fcs
    def __getattr__(self, key):
        def fc(*a,**b):
            return [getattr(fc, key)(*a,**b) for fc in self.fcs]
        return fc

pass
class DictCache(Base):
    '''
        用处：显存不够的时候，可以用本代码框架，代码会在forward和backward的时候自动把nets列表里需要计算的模型放入显存，不需要计算的放到内存，需要进行卷积训练的时候用处比较大（卷积计算显卡比CPU强太多）
        暂不支持多显卡，因为开发者的电脑没有多显卡
        DictCache实际提供两个功能：
            1，缓存功能：模型需要使用的时候放显存，不需要的时候放内存
            2，部分模型全放显存，部分模型全放内存，DictCache内部做传入数据的显存和内存的转换
            两个功能可以同时使用，即部分模型用的时候放显存，不用的时候放内存，部分模型全放显存，部分模型全放内存
        使用本框架需要使用者手动把模型拆成多个小模型，把小模型列表nets传入DictCache
        训练时大概有纯显卡二分之一到三分之一的性能，起码比CPU好，尤其是进行卷积计算，比cpu好太多
        训练完使用时的线性层则比CPU还慢，感觉没必要用，不过DictCache也提供了部分模型全放显存，部分模型全放内存的功能，显存不够的时候可以使用
        DictCache里的Dict意思是传入的模型列表nets会存字典里，Cache就是用内存作为显存的缓存
        代码实现原理是利用pytorch的几个勾子函数：
            model.register_forward_pre_hook会在模型forward之前调用
            model.register_forward_hook会在模型forward之后调用
            model.register_full_backward_hook会在模型反向梯度计算之后调用
            torch.autograd.graph.saved_tensors_hooks(hook_pack, hook_unpack):
                hook_pack会在模型forward的时候把之后反向梯度计算要用的tensor进行存储
                hook_unpack是在反向梯度计算的时候取回forward存储的tensor
        代码例子:
        code in buildz.gpuz.test.demo:

            from buildz.gpuz.torch import DictCache
            import torch
            from torch import nn,optim
            from torch.utils.data import DataLoader, Dataset
            class TestDataset(Dataset):
                def __init__(self, num, dims):
                    self.num = num
                    self.dims = dims
                    self.datas = torch.rand(num, dims)
                    self.targets = torch.rand(num, dims)
                def __len__(self):
                    return self.num
                def __getitem__(self, i):
                    return self.datas[i], self.targets[i]
            dims = 12
            dataset = TestDataset(30, dims)
            dataloader = DataLoader(dataset, 10)
            class MiniModel(nn.Module):
                def __init__(self, dims, mdims, num):
                    super().__init__()
                    nets = [nn.Linear(dims, mdims)]
                    nets += [nn.Linear(mdims,mdims) for i in range(num)]
                    nets.append(nn.Linear(mdims,dims))
                    self.nets = nn.Sequential(*nets)
                def forward(self, inputs):
                    return self.nets(inputs)
            models = [MiniModel(dims, 32, 3) for i in range(10)]
            opts = [optim.Adam(model.parameters(), lr=0.001) for model in models]
            #可以指定哪些模型全部放cuda或者全部放cpu
            cuda_models = [models[1],models[2]]
            cpu_models = [models[-1]]
            real_model = nn.Sequential(*models)
            loss_fn = torch.nn.MSELoss()
            def opt_step(net, opt):
                # 如果模型只是用来测试，不做训练，可以不传该函数，同时opts传入空就可以
                # 对模型的一些其他优化，可以写可以不写，主要是调用opt.step()进行当前小模型的模型训练
                # 另外，opt不一定就是优化函数，可以是任何数据，其只取决于创建DictCache的时候传入的opts是什么
                torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
                opt.step()
            cache = DictCache([torch.device('cuda'), torch.device('cpu')],models,opts,3,opt_step, [cuda_models,cpu_models])

            # 训练:
            [md.train() for md in models]
            for inputs,targets in dataloader:
                inputs,targets = inputs.cuda(),targets.cuda()
                [opt.zero_grad() for opt in opts]
                outs = cache.do_forward(lambda:real_model(inputs))
                loss = loss_fn(outs, targets)
                cache.do_backward(lambda: loss.backward())
                # opt.step()在do_backward里会自动调用
                print(loss.item())
            # 测试:
            inputs = torch.rand(1, dims).cuda()
            with torch.no_grad():
                outputs = cache.do_forward(lambda:real_model(inputs))
            print(outputs)
            
            # 对比不用DictCache的时候
            # 注意：模型放入DictCache之后会挂上勾子函数，不再用DictCache的时要先调用DictCache的remove方法
            cache.remove()
            full_opt = optim.Adam(real_model.parameters(), lr=0.001)
            # 训练:
            real_model.cuda()
            real_model.train()
            for inputs,targets in dataloader:
                inputs,targets = inputs.cuda(),targets.cuda()
                full_opt.zero_grad()
                outs = real_model(inputs)
                loss = loss_fn(outs, targets)
                loss.backward()
                torch.nn.utils.clip_grad_norm_(real_model.parameters(), max_norm=1.0)
                full_opt.step()
                print(loss.item())
            # 测试:
            inputs = torch.rand(1, dims).cuda()
            with torch.no_grad():
                outputs = real_model(inputs)
            print(outputs)
    '''
    def init(self, dvs, nets, opts=None, win_size=1, backward_deal=None, dvs_nets=None, fc_inputs_to=None):
        '''
            dvs[0]: 显卡设备，应该传入torch.device('cuda')
            dvs[1]: CPU设备，应该传入torch.device('cpu')
            dvs如果都传入torch.device('cpu')，则是完全CPU存储和计算
                如果都传入torch.device('cuda')，则是完全显卡存储和计算
            dvs[0]=dvs[1]的时候，可以不传列表，直接传一个dv就可以了
            nets: 小模型列表
            opts: 每个小模型对应的可选参数列表，如果要做训练，需要传入该项，opts需要和nets一一对应
            win_size: 显卡里最多存放多少个nets里的小模型，多线程的时候用处会比较大（后续开发），单线程的时候也是越大越好
            backward_deal: 梯度反向传播后的调用函数，训练的时候要传，本框架代码在第i个小模型梯度反向传播后会调用backward_deal(nets[i], opts[i])，里面加上梯度反向传播后需要做的代码，如果只是测试(eval)，不训练，该项可以不传
            dvs_nets: 不做缓存的小模型列表。如果要一部分做缓存，一部分不做，并且不做的那部分也放到nets和opts里了（方便统一调用backward_deal），则需要在这里单独再放一下，如果不做的部分没放在nets和opts里，则不需要传该参数
                dvs_nets = None or len(dvs_nets) == 1 or len(dvs_nets)==2
                dvs_nets[0]=None or [model1,model2,...] 固定在dvs[0]设备
                dvs_nets[1]=None or [model1,model2,...] 固定在dvs[1]设备，如果len(dvs_nets)==1则没有这项
        '''
        if fc_inputs_to is None:
            fc_inputs_to = self.inputs_to
        self.fc_inputs_to = fc_inputs_to
        if type(dvs) not in (list,tuple):
            dvs = [dvs]
        self.gdv = dvs[0]
        self.cdv = dvs[-1]
        self.dvs = dvs
        if dvs_nets is None:
            dvs_nets = [[]]
        if type(dvs_nets) not in (list, tuple):
            dvs_nets = [dvs_nets]
        dvs_nets = [k if k is not None else [] for k in dvs_nets]
        ids_nocache = set()
        fcs = []
        for i in range(len(dvs_nets)):
            for net in dvs_nets[i]:
                fc = net.register_forward_pre_hook(self.hook_forward_pre_tensor(i))
                fcs.append(fc)
                ids_nocache.add(id(net))
        self.ids_nocache = ids_nocache
        self.dvs_nets = dvs_nets
        fcs += [net.register_forward_pre_hook(self.hook_forward_pre) for net in nets if id(net) not in ids_nocache]
        fcs+=[net.register_forward_hook(self.hook_forward) for net in nets if id(net) not in ids_nocache]
        net = nets[0]
        if hasattr(net, "register_full_backward_hook"):
            fcs+=[net.register_full_backward_hook(self.hook_backward) for net in nets]
        else:
            fcs+=[net.register_backward_hook(self.hook_backward) for net in nets]
        self.src_nets = nets
        self.fcs = fcs
        self.nets = {id(net):net for net in nets}
        self.ctxs = {id(net):[] for net in nets}
        if opts is not None:
            opts = {id(net): opt for net,opt in zip(nets, opts)}
        self.opts = opts
        self.pools = []
        self.win_size = win_size
        self.backward_deal = backward_deal
        self.nears = {id(net):[-1,-1] for net in nets}
        self.curr = -1
        self.done_backward = -1
        self.no_caches = True
    def remove(self):
        [fc.remove() for fc in self.fcs]
    def hook_forward_pre_tensor(self, i):
        def hook(model, ins):
            model.to(self.dvs[i])
            ins = self.fc_inputs_to(ins, self.dvs[i])
            return ins
        return hook
    @staticmethod
    def inputs_to(inputs, dv):
        if isinstance(inputs, torch.Tensor):
            return inputs.to(dv)
        inputs = tuple([k.to(dv) for k in inputs])
        return inputs
    def ctxs_to(self, i, dv):
        if dv is None:
            self.ctxs[i] = []
        else:
            self.ctxs[i] = [k.to(dv) for k in self.ctxs[i]]
    def copy_backward(self, nid):
        for c_id in self.pools:
            self.nets[c_id].to(self.cdv)
            self.ctxs_to(c_id, None)
        self.pools = []
        self.nets[nid].to(self.gdv)
        self.ctxs_to(nid, self.gdv)
        self.pools.append(nid)
        next_id = nid
        for i in range(self.win_size-1):
            next_id = self.nears[next_id][1]
            if next_id<0:
                break
            self.nets[next_id].to(self.gdv)
            self.ctxs_to(next_id, self.gdv)
            self.pools.append(next_id)
    def copy_forward(self, nid, model):
        for c_id in self.pools:
            self.nets[c_id].to(self.cdv)
            self.ctxs_to(c_id, self.cdv)
        self.pools = []
        model.to(self.gdv)
        self.pools.append(nid)
        next_id = nid
        for i in range(self.win_size-1):
            next_id = self.nears[next_id][0]
            if next_id<0:
                break
            self.nets[next_id].to(self.gdv)
            self.pools.append(next_id)
    def hook_forward(self, model, ins, outs):
        self.no_caches = True
    def hook_forward_pre(self, model, ins):
        self.no_caches = False
        nid = id(model)
        #print(f"hook_forward_pre {nid}")
        if nid not in self.pools:
            self.copy_forward(nid, model)
        if self.curr>=0:
            self.nears[self.curr][0] = nid
        self.curr = nid
        return self.fc_inputs_to(ins, self.gdv)
    def do_forward(self, fc):
        #print("FWD")
        self.no_caches = True
        self.ctxs = {k:[] for k in self.nets}
        with torch.autograd.graph.saved_tensors_hooks(self.hook_pack, self.hook_unpack):
            rst = fc()
        #print("DFWD")
        return rst.to(self.gdv)
    def wrap_backward_deal(self, net_id):
        if self.backward_deal is None:
            return
        self.backward_deal(self.nets[net_id], self.opts[net_id])
    def hook_backward(self, model, grad_ins, grad_outs):
        nid = id(model)
        #print(f"hook_backward {nid}")
        self.wrap_backward_deal(nid)
    def hook_pack(self, dt):
        #print(f"pack")
        if self.no_caches:
            # 不做缓存，数据不处理
            return -1, dt
        index = self.curr
        # forward时候为了后面计算梯度存的缓存，放到列表里方便转cpu和gpu
        self.ctxs[index].append(dt)
        return index, len(self.ctxs[index])-1
    def hook_unpack(self, x):
        nid = x[0]
        #print(f"unpack {nid}")
        if nid<0:
            return x[1]
        if nid not in self.pools:
            self.copy_backward(nid)
        dt = self.ctxs[nid][x[1]]
        if self.curr>=0:
            self.nears[self.curr][1] = nid
        self.curr = nid
        return dt
    def do_backward(self, fc):
        self.curr = -1
        fc()




from .. import base
from .. import itemz
class ListDeal(base.BaseDeal):
    def init(self, left, right, spt):
        self.left = left
        self.right = right
        self.spt = spt
    def deal(self, obj, conf):
        if not obj.check(is_list=1):
            return None
        arr = obj.val
        rs = None
        rst = []
        deep = 0
        for i in range(len(arr)):
            obj = arr[i]
            _val = obj.val
            deep = max(obj.deep+1, deep)
            if i < len(arr)-1:
                _val += conf.s(self.spt)
            rst.append(_val)
        if not conf.check(format=1) or conf.get(deep=0)+1>=deep:
            spc = conf.s(" ")*conf.get(prev=0)
            rst = spc.join(rst)
        else:
            rst = [self.fmt(k, conf.get(line=4), conf.get(spc=' ')) for k in rst]
            rst = conf.s("\n").join(rst)
            rst = conf.s("\n")+rst+conf.s("\n")
        rst = conf.s(self.left)+rst+conf.s(self.right)
        return itemz.ShowItem(rst, deep, is_list=1, is_done=1)

pass 



            

                


#
import torch
from torch import nn
import threading as th
'''
注：
代码有误，原本以为register_full_backward_hook的勾子函数是在模型反向梯度计算之前调用的，后来发现是反向梯度计算之后调用的，本来本代码使用局限就比较多（要求拆分的模型线性连接），直接重写了个buildz.gpuz.torch.dict_middle作为代替，本代码暂时先放着，后续会做修改或者删除
'''
class SeqCache:
    '''
        用处：显存不够，同时模型可以拆成多个小模型线性连接的时候，可以用本代码，本代码会在forward和backward的时候自动把小的多层网络轮流放到gpu里计算，计算完再转cpu里
        需要使用者手动将多层网络拆分成多个更小一点的多层网络
        测试代码见test_moddle_conv1.py
        大概有纯显卡二分之一到三分之一的性能，起码比cpu好，尤其是进行卷积计算，比cpu好太多
        代码例子:
        
        from buildz.gpuz.torch import CacheModel
        from torch import nn,optim
        model1 = nn.Sequential(*[nn.Linear(1024,1024) for i in range(10)])
        model2 = nn.Sequential(*[nn.Linear(1024,1024) for i in range(10)])
        model3 = nn.Sequential(*[nn.Linear(1024,1024) for i in range(10)])
        opt1 = optim.Adam(model1.parameters(), lr=0.001)
        opt2 = optim.Adam(model2.parameters(), lr=0.001)
        opt3 = optim.Adam(model3.parameters(), lr=0.001)
        models = [model1,model2,model3]
        opts = [opt1,opt2,opt3]
        loss_fn = torch.nn.MSELoss()
        def opt_step(net, opt):
            # 如果模型只是用来测试，不做训练，可以不传该函数，同时opts传入空就可以
            # 对模型的一些其他优化，可以写可以不写，主要是调用opt.step()进行当前小模型的模型训练
            # torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
            opt.step()
        cmodel = CacheModel(torch.device('cuda'), torch.device('cpu'),models,opts,3,opt_step)

        # 训练:
        [md.train() for md in models]
        for inputs,targets in dataloader: #批量数据集，这个自己实现
            [opt.zero_grad() for opt in opts]
            outs = cmodel.do_forward(inputs)
            loss = loss_fn(outs, targets)
            cmodel.do_backward(lambda: loss.backward())
            # opt.step()在do_backward里会自动调用
            print(loss.item())

        # 测试:
        with torch.no_grad():
            outputs = cmodel.do_forward(inputs)
        print(outputs)
    '''
    def __init__(self, gdv, cdv, nets, opts, win_size = 1, backward_deal = None):
        '''
            gdv: 显卡设备，应该传入torch.device('cuda')
            cdv: CPU设备，应该传入torch.device('cpu')
            gdv和cdv如果都传入torch.device('cpu')，则是完全CPU存储和计算
                如果都传入torch.device('cuda')，则是完全显卡存储和计算
        '''
        self.gdv = gdv
        self.cdv = cdv
        # event和condition未使用，本来打算做成多线程，但python线程一次只能有一个在运行，抢占有点严重，待修改
        self.event = th.Event()
        self.condition=th.Condition()
        [net.register_full_backward_hook(self.hook_backward) for net in nets]
        self.nets = nets
        self.size = len(nets)
        self.ctxs = [[] for i in range(self.size)]
        self.size_1 = self.size-1
        self.opts = opts
        self.win_size = win_size
        self.backward_deal = backward_deal
        self.base = -1
        self.curr = 0
        self.last = -1
        self.running = False
    def hook_pack(self, dt):
        # forward时候为了后面计算梯度存的缓存，放到列表里方便转cpu和gpu
        self.ctxs[self.curr].append(dt)
        return len(self.ctxs[self.curr])-1
    def hook_unpack(self, x):
        dt = self.ctxs[self.curr][x]
        return dt
    def nfc(self, fc, *a,**b):
        [getattr(net, fc)(*a,**b) for net in self.nets]
    def reset(self):
        for i in range(self.base,self.last+1):
            self.nets[i].to(self.cdv)
            self.ctxs_to(i, self.cdv)
        self.base,self.last=-1,-1
    def ctxs_to(self, i, dv):
        if dv is None:
            self.ctxs[i] = []
        else:
            self.ctxs[i] = [k.to(dv) for k in self.ctxs[i]]
    def copy_backward(self):
        if self.last<self.curr:
            self.reset()
        if self.base==0:
            return False
        if self.last<0:
            self.nets[self.size_1].to(self.gdv)
            self.last = self.size_1
            self.base = self.size_1
        diff = self.win_size-(self.last-self.base+1)
        diff = min(diff, self.base)
        for i in range(diff):
            self.nets[self.base-1].to(self.gdv)
            self.ctxs_to(self.base-1, self.gdv)
            self.base-=1
        rels = self.last-self.curr
        for i in range(rels):
            self.nets[self.last].to(self.cdv)
            self.ctxs_to(self.last, None)
            self.last-=1
        return True
    def copy_forward(self):
        if self.base>self.curr:
            self.reset()
        if self.last==self.size_1:
            return False
        if self.base<0:
            self.nets[0].to(self.gdv)
            self.base=0
            self.last=0
        diff = self.win_size-(self.last-self.base+1)
        diff = min(diff, self.size_1-self.last)
        for i in range(diff):
            self.nets[self.last+1].to(self.gdv)
            self.last+=1
        rels = self.curr-self.base
        for i in range(rels):
            self.nets[self.base].to(self.cdv)
            self.ctxs_to(self.base, self.cdv)
            self.base+=1
        return True
    def th_copy_forward(self):
        while self.copy_forward():
            self.event.set()
        self.running = False
    def th_copy_backward(self):
        while self.copy_backward():
            self.event.set()
    def wait(self):
        with self.condition:
            self.condition.notify()
        self.event.wait()
    def do_forward(self, inputs):
        # while self.running:
        #     import time
        #     time.sleep(0.01)
        # t = th.Thread(target=self.th_copy_forward, daemon=True)
        # self.running = True
        # t.start()
        self.ctxs = [[] for i in range(self.size)]
        with torch.autograd.graph.saved_tensors_hooks(self.hook_pack, self.hook_unpack):
            rst = self.forward(inputs)
        return rst
    def forward(self, inputs):
        for self.curr in range(len(self.nets)):
            while not (self.base<=self.curr<=self.last):
                self.copy_forward()
                #self.wait()
            inputs = self.nets[self.curr](inputs)
        return inputs
    def wrap_backward_deal(self, i):
        if self.backward_deal is None:
            return
        try:
            self.backward_deal(self.nets[i], self.opts[i])
        finally:
            pass
    def hook_backward(self, model, ins, outs):
        if self.backward_curr<self.size_1:
            self.wrap_backward_deal(self.backward_curr+1)
        self.curr = self.backward_curr
        while not (self.base<=self.backward_curr<=self.last):
            #self.wait()
            self.copy_backward()
        self.backward_curr-=1
    def do_backward(self, fc):
        self.backward_curr=self.curr
        # t = th.Thread(target=self.th_copy_backward,daemon=True)
        # t.start()
        fc()
        self.wrap_backward_deal(0)

pass




from .. import base
from .. import itemz
import re
import json
class ValDeal(base.BaseDeal):
    def init(self, fc):
        self.fc = fc
    def deal(self, mg, obj, up_height, heights, conf, deep, top_not_format):
        val = conf.s(self.fc(obj))
        val = self.prev_spc(conf, deep, up_height, top_not_format)+val
        return val

pass 



            

                



from . import mapz
from buildz import xf
from ..base import Base
import os
def dzkeys(key, spt):
    if key is None:
        return []
    # return mapz.keys(key, spt)
    if type(key)==str:
        key = key.split(spt)
    if type(key) not in (list, tuple):
        key = [key]
    return key

class BindKey(Base):
    def sub(self, key):
        return BindKey(key, False, self, self.obj)
    def init(self, key, abs=True, up=None, conf=None):
        self.key = key
        self.abs = abs
        self.spt = spt
        self.up = up
        self.obj = conf
    def bind(self, conf):
        self.obj=conf
    def get(self, conf=None, default=None):
        conf = self.conf(conf)
        if self.up:
            conf = self.up(conf)
        elif self.abs:
            conf = conf()
        return conf.get(self.key, default)
    def conf(self, obj):
        obj = obj or self.obj
        assert obj is not None
        return obj
    def set(self, val, conf=None):
        conf = self.conf(conf)
        if self.up:
            conf = self.up(conf)
        elif self.abs:
            conf = conf()
        return conf.set(self.key, val)
    def call(self, conf):
        conf = self.conf(conf)
        if self.up:
            conf = self.up(conf)
        elif self.abs:
            conf = conf()
        return conf(self.key)

pass
'''

实际数据：
src: Conf对象，最底层，只读
conf: dict数据，实际数据，可读可写
history: dict里是list，每个key在push前，把当前key对应的数据存入history中，后面pop的时候，删除当前数据，把之前的数据重新更新
_links: link对象，搜索方式是搜最长的有link的节点，返回link目标key拼接剩余key作为key的取值
link对象：[{子集}, 当前节点link目标, 是否有link(1是，0否)]，初始化为[{}, None, 0]


'''
class Conf(Base):
    @staticmethod
    def bind_key(key, abs=True, up = None):
        return BindKey(key, abs, up)
    def val(self):
        return self.get_conf()
    def get_type(self):
        return type(self.val())
    def has_val(self):
        obj = self.root or self
        return obj.has(self.domain)
    def top(self, domain = None, loop=0, link=0):
        root = self.root or self
        if domain is not None:
            root = root(domain, loop, link)
        return root
    def ltop(self, domain=None, loop=-1,link=0):
        return self.top(domain,loop,link)
        # return self.root or self
    def get_conf(self):
        if self.domain:
            key = self.domain
        obj = self.root or self
        if self.domain:
            return obj._get(self.domain)
        return obj.conf
    def str(self):
        return str(self.get_conf())
    def l(self, domain=None, loop=-1,link=0):
        return self(domain,loop,link)
    def call(self, domain=None, loop=0, link=0):
        if domain is None:
            return self.top()
        if self.domain:
            domain = self.domain+self.spt+domain
        obj = self.root or self
        if loop!=0:
            val, find = obj.hget(domain,link=link)
            while loop!=0 and find and type(val)==str:
                domain = val
                val,find = obj.hget(domain,link=link)
                loop-=1
        return Conf(self.spt, self.spts, domain, obj)
    def init(self, spt='.', spts=',', domain=None, root = None, src = None):
        self.spt = spt
        self.spts = spts
        self.domain = domain
        self.root = root
        self.src = src
        if root is None:
            self.conf = {}
            self.history = {}
            self._links = [{},None,0]
        self.dr_bind('_get', 'get')
        self.dr_bind('_hget', 'hget')
        self.dr_bind('_lget', 'lget')
        self.dr_bind('_lhget', 'lhget')
        self.dr_bind('_set', 'set')
        self.dr_bind('_has', 'has')
        self.dr_bind('_remove', 'remove')
        self.dr_bind('_link', 'link')
        self.dr_bind('_unlink', 'unlink')
        self.dr_bind('_push', 'push')
        self.dr_bind('_pop', 'pop')
        self.dr_bind('_stack_set', 'stack_set')
        self.dr_bind('_stack_unset', 'stack_unset')
        self.fcs_bind('get', 'gets', False, True)
        self.fcs_bind('lget', 'lgets', False, True)
        self.fcs_bind('set', 'sets', True)
        self.fcs_bind('remove', 'removes')
        self.fcs_bind('push', 'pushs', True)
        #self.fcs_bind('pop', 'pops')
        self.fcs_bind('stack_set', 'stack_sets', True)
        self.fcs_bind('stack_unset', 'stack_unsets')
        self.fcs_bind('link', 'links', True)
        self.fcs_bind('unlink', 'unlinks')
        for name,rename in zip("stack_set,stack_unset,stack_sets,stack_unsets".split(","), "tmp_set,tmp_unset,tmp_sets,tmp_unsets".split(',')):
            setattr(self, rename, getattr(self, name))
        self.have_all = self.has_all
    def clean(self):
        obj = self.root or self
        obj.conf = {}
        obj.history = {}
        obj._links = [{}, None, 0]
        return self
    def dkey(self, key):
        if self.domain:
            key = self.domain+self.spt+key
        return key
    def update(self, conf, flush = 1, replace=1, visit_list=0):
        if self.domain:
            ks = dzkeys(self.domain, self.spt)
            tmp = {}
            mapz.dset(tmp, ks, conf)
            conf = tmp
        if self.root:
            return self.root.update(conf, flush, replace, visit_list)
        if flush:
            conf = xf.flush_maps(conf, lambda x:x.split(self.spt) if type(x)==str else [x], visit_list)
        xf.fill(conf, self.conf, replace=replace)
        return self
    def dr_bind(self, fn, wfn):
        def wfc(key,*a,**b):
            key = self.dkey(key)
            obj = self.root or self
            fc = getattr(obj, fn)
            return fc(key, *a, **b)
        setattr(self, wfn, wfc)
    def fcs_bind(self, fn, wfn, align=False, null_default= False):
        def wfc(keys, *objs, **maps):
            keys = self.spts_ks(keys)
            fc = getattr(self, fn)
            rst = []
            for i in range(len(keys)):
                if i<len(objs):
                    val = fc(keys[i], objs[i], **maps)
                else:
                    if align:
                        raise Exception(f"not val[{i}]")
                    if null_default:
                        val = fc(keys[i], None, **maps)
                    else:
                        val = fc(keys[i], **maps)
                rst.append(val)
            return rst
        setattr(self, wfn, wfc)
    def _stack_set(self, key, value, flush = 1, update=0):
        return self._push(key, value, flush, update, 1)
    def _stack_unset(self, key):
        return self._pop(key, 1)
    def _push(self, key, value, flush = 1, update=0, clean_history = 0):
        keys = dzkeys(key, self.spt)
        val, find = mapz.dget(self.conf, keys)
        val = mapz.deep_clone(val)
        if clean_history or key not in self.history:
            self.history[key] = []
        self.history[key].append([val, find, update])
        if flush and type(value)==dict:
            value = xf.flush_maps(value, lambda x:x.split(self.spt) if type(x)==str else [x], 0)
        if type(value)==dict and update:
            self(key).update(value,flush=0)
        else:
            self._set(key, value)
    def _pop(self, key, clean_history = 0):
        if key not in self.history:
            return False
        lst = self.history[key]
        if len(lst)==0:
            return False
        rst = lst.pop(-1)
        if clean_history:
            self.history[key] = []
        if not rst[1]:
            self._remove(key)
            return True
        self._set(key, rst[0])
        return True
    def pops(self, keys, *a, **b):
        keys = self.spts_ks(keys)
        keys.reverse()
        for key in keys:
            self.pop(key)
    def with_push(self, key, *a, **b):
        self.push(key, *a, **b)
        def out():
            self.pop(key)
        return pyz.with_out(out)
    def with_pushs(self, keys, *a, **b):
        self.pushs(keys, *a, **b)
        def out():
            self.pops(keys)
        return pyz.with_out(out)
    def ld_get(self, obj, key, fc_init=None):
        keys = dzkeys(key, self.spt)
        for key in keys:
            if key not in obj[0]:
                if fc_init is None:
                    return None
                obj[0][key] = fc_init()
            obj = obj[0][key]
        return obj
    def ld_visit(self, obj, key, fc):
        keys = dzkeys(key, self.spt)
        deep = 0
        fc(obj, deep, len(keys))
        for key in keys:
            if key not in obj[0]:
                break
            deep+=1
            obj = obj[0][key]
            fc(obj, deep, len(keys))
    def _link(self, src, target):
        #print(f"[LINK] {src}->{target}")
        links = self.ld_get(self._links, src, lambda :[{}, None, 0])
        links[1] = target
        links[2] =1
    def _unlink(self, key):
        links = self.ld_get(self._links, key)
        if links is None:
            return False
        links[1] = None
        links[2] = 0
        return True
    def link_match(self, keys):
        obj = self.root or self
        rst = [[None,0,0]]
        def fc_match(val, deep, size):
            if not val[2]:
                return
            rst[0] = val[1], val[2], deep
        self.ld_visit(obj._links, keys, fc_match)
        return rst[0]
    def _set(self, key, val):
        keys = dzkeys(key, self.spt)
        mapz.dset(self.conf, keys, val)
    def _lhget(self, key, default=None, loop=-1, link=-1):
        a,b = self._hget(key, default, link)
        bak = a, b
        while b and type(a)==str and loop!=0:
            a,b = self._hget(a, default, link)
            if b:
                bak = a,b
            loop-=1
        return bak
    def key(self, ks):
        return self.spt.join(ks)
    def src_hget(self, key, default, link=-1):
        if self.src is None:
            return default, 0
        return self.src.hget(key, default, loop, link)
    def _lget(self, key, default=None, loop=-1, link=-1):
        return self._lhget(key, default, loop, link)[0]
    def _hget(self, key, default=None, link=-1):
        #print(f"_hget: {key}, {default}, {link}")
        keys = dzkeys(key, self.spt)
        val, find = mapz.dget(self.conf, keys, default)
        if find or link==0:
            if not find:
                val, find = self.src_hget(key, default, link)
            return val, find
        lnk, has_lnk, deep = self.link_match(keys)
        if not has_lnk:
            return self.src_hget(key, default, link)
            #return val, find
        keys = keys[deep:]
        key = self.spt.join(keys)
        if lnk is not None:
            key = lnk+self.spt+key
        if link>0:
            link-=1
        return self._hget(key, default, link)
    def _get(self, key, default=None, link=-1):
        return self._hget(key, default,link)[0]
        keys = dzkeys(key, self.spt)
        return mapz.dget(self.conf, keys, default)[0]
    def _remove(self, key):
        # TODO
        keys = dzkeys(key, self.spt)
        return mapz.dremove(self.conf, keys)
    def _has(self, key, link=-1):
        return self._hget(key, None, link)[1]
        keys = dzkeys(key, self.spt)
        return mapz.dhas(self.conf, keys)
    def spts_ks(self, keys):
        keys = dzkeys(keys, self.spts)
        keys = [k.strip() if type(k) == str else k for k in keys]
        return keys
    def g(self, **maps):
        rst = [self.get(k, v) for k,v in maps.items()]
        if len(rst)==1:
            rst = rst[0]
        return rst
    def s(self, **maps):
        [self.set(k,v) for k,v in maps.items()]
    def has_all(self, keys, link = -1):
        keys = self.spts_ks(keys)
        rst = [1-self.has(key, link) for key in keys]
        return sum(rst)==0
    def has_any(self, keys, link=0):
        keys = self.spts_ks(keys)
        for key in keys:
            if self.has(key, link):
                return True
        return False

pass

class ArgsConf(Base):
    def init(self, conf, args):
        self.conf =conf
        self.args =args
    def __getattr__(self, key):
        return getattr(self.conf,key)

pass


#

from buildz import xf

s = r"""
a=0
b=1
c={
    a=2,
    d=3
    e=4
    f=5
    g=6
}
h=[
    {a=0,b=1,c=2}
    asdf
    'haha'
    [0,0,0,a=0,b=1,c=2]
]
"""
obj = xf.loadx(s)
print("obj")
print(obj)

rst = xf.dumpx(obj)
print("rst:")
print(rst)
print("\n")
rst = xf.dumpx(obj,format=1)
print("rst:")
print(rst)
print("\n")
rst = xf.dumpx(obj,format=1, not_format_height=2)
print("rst:")
print(rst)
print("\n")
rst = xf.dumpx(obj,format=1, not_format_height=2,json_format=1)
print("rst:")
print(rst)
print("\n")

obj = xf.loadx(rst)
print(obj)


#
from buildz import Base, xf,pyz
class Build(Base):
    def init(self, *binds):
        self.builder = None
        self.binds = binds
    def bind(self, builder):
        self.builder = builder
        for obj in self.binds:
            obj.bind(builder)
    def call(self, conf):
        assert 0, 'not impl'
class Builder(Base):
    def init(self, fc = None, key_id = "id", default_key = "main"):
        self.vars = {}
        self.confs = {}
        self.key_id = key_id
        self.df_key = default_key
        self.set_fc(fc)
    def set_fc(self, fc):
        fc.bind(self)
        self.fc = fc
        return self
    def var(self, key, obj):
        self.vars[key] = obj
        return self
    def get_var(self, key):
        return self.vars[key]
    def get_conf(self, key):
        if type(key) == dict:
            conf = key
        else:
            if key not in self.confs:
                return self.get_var(key)
            conf = self.confs[key]
        return self.fc(conf)
    def conf(self, data):
        if type(data) == dict:
            tp = xf.g(data, type=None)
            if tp is not None and type(tp) not in (list, tuple, dict):
                if self.key_id not in data:
                    data[self.key_id] = self.df_key
                data = [data]
        if type(data) in (list, tuple):
            rst = {}
            for it in data:
                if self.key_id not in it:
                    continue
                id = it[self.key_id]
                rst[id] = it
            data = rst
        self.confs = data
        return self
    def call(self, data, key=None):
        key = pyz.nnull(key, self.df_key)
        return self.conf(data).get_conf(key)



#

from .middle_base import MiddleBase
import torch
from buildz import logz
log = logz.simple("middle_cache.log")("GPUZ_CACHE")
log.show("debug", 0)
class MiddleCache(MiddleBase):
    '''
        nets: 模型列表
        win_size: gpu设备可存放的模型数量
        deal_after_backup: fc(model) 或 None, 在模型进行backward后调用
        cal_dv: 计算设备，默认cuda
        cache_dv: 缓存设备，默认cpu
        fc_inputs_to: fc(inputs, dv) 或 None, 需要把输入参数转换成设备dv对应的数据时候使用，默认输入参数是Tensor的tuple或者就是Tensor
        cal_nets: 模型列表，nets中的一部分，不进行两个设备转换，而是一直放在cal_dv上的模型
        cache_nets: 模型列表，nets中的一部分，不进行两个设备转换，而是一直放在cache_dv上的模型
        preload: 预加载，把模型加载到计算设备时，把后面的模型也加载到计算设备上（已经加载到计算设备上的放缓存设备中）
    '''
    def init(self, nets, win_size=1, deal_after_backup=None, cal_dv='cuda', cache_dv='cpu', fc_inputs_to = None, cal_nets = [], cache_nets = [], preload = False):
        cache_dv = torch.device(cache_dv) if type(cache_dv)==str else cache_dv
        cal_dv = torch.device(cal_dv) if type(cal_dv)==str else cal_dv
        [net.to(cache_dv) for net in cache_nets]
        [net.to(cal_dv) for net in cal_nets]
        self.src = nets
        static_dv = {}
        static_dv.update({id(net):cache_dv for net in cache_nets})
        static_dv.update({id(net):cal_dv for net in cal_nets})
        self.static_dv = static_dv
        self.nets = {id(net):net for net in nets if id(net) not in self.static_dv}
        # TEST CODE
        # self.indexes = {}
        # for i in range(len(nets)):
        #     self.indexes[id(nets[i])] = i
        # DONE
        self.cache_dv = cache_dv
        self.cal_dv = cal_dv
        self.hook(nets)
        # 
        self.datas = {id(net):[] for net in nets}
        self.cal_ids = []
        self.deal_after_backup = deal_after_backup
        if fc_inputs_to is None:
            fc_inputs_to = self.list_inputs_to
        self.inputs_to = fc_inputs_to
        self.nexts = {}
        self.prevs = {}
        self.curr = -1
        #self.caches = set()
        self.mark_static = None
        self.win_size = win_size
        self.preload = preload
    def train(self):
        [net.train() for net in self.src]
    def eval(self):
        [net.eval() for net in self.src]
    @staticmethod
    def list_inputs_to(datas, dv):
        if isinstance(datas, torch.Tensor):
            return datas.to(dv)
        # assert datas is not None, "inputs datas is null"
        inputs = tuple([k.to(dv) for k in datas])
        return inputs
    @staticmethod
    def deep_inputs_to(datas, dv):
        '''
            数据比较复杂用这个，会递归搜索list，tuple和dict病做转换
        '''
        log.debug(f"deep_inputs_to")
        if isinstance(datas, torch.Tensor):
            return datas.to(dv)
        if type(datas) in (list, tuple):
            rst = [MiddleCache.deep_inputs_to(k, dv) for k in datas]
            if type(datas)==tuple:
                rst = tuple(rst)
            return rst
        if type(datas) == dict:
            rst = {k:MiddleCache.deep_inputs_to(v, dv) for k,v in datas.items()}
            return rst
        assert 0, f"undealable datas type: {type(datas)}"
    def before_forward(self):
        log.debug(f"before_forward")
        self.curr = -1
        #self.caches = set()
        for mid in self.datas:
            self.datas[mid] = []
    def after_forward(self):
        pass
    def before_backward(self):
        log.debug(f"before_backward")
        self.curr = -1
    def backupable(self, model):
        mid = id(model)
        return mid in self.cal_ids or mid in self.static_dv
        #return mid not in self.caches
    def after_backward(self):
        pass
    def to_dv(self, mid, dv):
        self.nets[mid].to(dv)
        self.datas[mid] = [self.inputs_to(k, dv) if k is not None else k for k in self.datas[mid]]
    def to_cal(self, mid, pop = 0):
        log.debug(f"to_cal")
        if mid in self.cal_ids:
            return 0
        if len(self.cal_ids)>=self.win_size:
            xid = self.cal_ids.pop(pop)
            #print(f"model_{self.indexes[xid]} to CPU device")
            self.to_dv(xid, self.cache_dv)
            #self.caches.add(xid)
        #print(f"model_{self.indexes[mid]} to CUDA device")
        self.to_dv(mid, self.cal_dv)
        # if mid in self.caches:
        #     self.caches.remove(mid)
        if pop==-1:
            self.cal_ids = [mid]+self.cal_ids
        else:
            self.cal_ids.append(mid)
        return 1
    def hook_forward_before(self, model, ins):
        mid = id(model)
        log.debug(f"hook_forward_before", model, id(model))
        self.mark_static = mid in self.static_dv
        if self.mark_static:
            # assert ins is not None, "hook_forward_before A"
            return self.inputs_to(ins, self.static_dv[mid])
        prev = self.curr
        if prev >= 0:
            self.nexts[prev] = mid
            if mid not in self.prevs:
                self.prevs[mid] = prev
        self.curr = mid
        self.loop_to_cal(mid, 0, self.nexts)
        # assert ins is not None, "hook_forward_before B"
        return self.inputs_to(ins, self.cal_dv)
    def hook_forward_after(self, model):
        log.debug(f"hook_forward_after", model, id(model))
        pass
    def hook_backward_after(self, model):
        log.debug(f"hook_backward_after", model, id(model))
        #print(f"hook_backward_after", model, id(model))
        assert self.backupable(model)
        mid = id(model)
        if self.deal_after_backup is not None:
            self.deal_after_backup(model)
        if id in self.static_dv:
            return
        after = self.curr
        if after>=0:
            self.prevs[after] = mid
        self.curr = mid
    def tensor_save(self, tuple_data):
        log.debug(f"tensor_save")
        if self.mark_static:
            return -1, tuple_data
        index = len(self.datas[self.curr])
        self.datas[self.curr].append(tuple_data)
        log.debug(f"done tensor_save", self.curr, index)
        return self.curr, index
    def tensor_load(self, obj):
        log.debug(f"tensor_load", obj)
        if obj[0]<0:
            return obj[1]
        curr, index = obj
        self.loop_to_cal(curr, -1, self.prevs)
        tuple_data = self.datas[curr][index]
        # assert tuple_data is not None
        self.datas[curr][index] = None
        return tuple_data
    def loop_to_cal(self, mid, pop, nexts):
        log.debug(f"loop_to_cal")
        if mid in self.cal_ids:
            return
        self.to_cal(mid, pop)
        if not self.preload:
            return
        count = 1
        while count<self.win_size and mid in nexts:
            mid = nexts[mid]
            if mid in self.cal_ids:
                break
            self.to_cal(mid, pop)
            count+=1
            
pass


from .timez import Clock,clock,timecost,showcost
import time
class Timer:
    def __init__(self):
        self.curr = None
    def start(self):
        self.curr = time.time()
    def __call__(self, reset=False):
        curr = time.time()
        sec = curr-self.curr
        if reset:
            self.curr = curr
        return sec

pass


#
from . import jsc, defs, dm, rn, fcs
import re,sys, os
from buildz import fz, dz, xf, argx, pyz, Base
fetch =argx.Fetch(*xf.loads("[fp,ofp,left,right,spt],{f:fp,o:ofp,l:left,r:right,s:spt}"))
def test():
    conf = fetch()
    fp, ofp = dz.g(conf, fp=None,ofp=None)
    if ofp is None:
        i = fp.rfind(".")
        if i>0:
            ofp = fp[:i]+".java"
        else:
            ofp = fp
    left, right, spt = dz.g(conf, left='<<', right='>>', spt = '||')
    deals = jsc.DealCodes(left, right, spt)
    deals.add("default", defs.default)
    deals.add("domain", dm.domains)
    deals.add("rename", rn.renames)
    deals.add("methods", fcs.fcs)
    s = fz.read(fp).decode("utf-8")
    rs = deals.update(s).encode("utf-8")
    fz.write(rs, ofp)
    print("done to", ofp)

pass
fetch =argx.Fetch(*xf.loads("[dp,odp,left,right,spt],{d:dp,o:odp,l:left,r:right,s:spt}"))
def tests():
    conf = fetch()
    dp, odp = dz.g(conf, dp=None,odp=None)
    dp = dp or '.'
    odp = odp or dp
    print(f"dp:", dp, sys.argv)
    fs = fz.search(dp, '.*\.javap')
    print(fs)
    left, right, spt = dz.g(conf, left='<<', right='>>', spt = '||')
    deals = jsc.DealCodes(left, right, spt)
    deals.add("default", defs.defaults)
    deals.add("domain", dm.domains)
    deals.add("rename", rn.renames)
    deals.add("methods", fcs.fcs)
    for fp in fs:
        fn = fp
        if len(dp)>0:
            fn = fp[len(dp):]
            if fn[0] in "\\/":
                fn = fn[1:]
        #print(f"[TESTZ] fn:]: {fn}")
        #print(f"[TESTZ] odp: {odp}")
        ofp = os.path.join(odp, fn)
        i = ofp.rfind(".")
        if i>0:
            ofp = ofp[:i]+".java"
        else:
            ofp = ofp+".java"
        #print(f"[TESTZ] fp: {fp}")
        s = fz.read(fp).decode("utf-8")
        rs = deals.update(s).encode("utf-8")
        #print(f"[TESTZ] ofp: {ofp}")
        fz.write(rs, ofp)
    print("done to", odp)



pyz.lc(locals(), tests)

"""
python -m main Conf.javap
python -m main Utils.javap

python -m buildz.jv.jvp.main ../confz

"""


#
from .datas import Datas
from .dataset import Dataset
from .confs import Confs
class Encapes(Datas):
    def init(self, ns=None, id=None, dts=None, unit=None):
        super().init(ns, id, dts)
        self.confs = None
        self.deals = None
        self.deal_key = None
        self.unit = unit
        self.build()
    def build(self):
        if self.unit is None or self.confs is not None:
            return
        self.confs = self.unit.confs
        self.deals = self.unit.deals
        self.deal_key = self.unit.deal_key
    def make(self, key, ns=None, id=None):
        if self.confs is None or self.deals is None or self.deal_key is None:
            return 0,0,0
        conf, ctag, cfind = self.confs.tget(key, ns, id, False)
        if not cfind:
            return 0,0,0
        conf, deal_ns, cid = conf
        encape, conf, upd = self.make_conf(conf, deal_ns, id)
        if upd:
            self.confs.set(key, conf, ctag)
        self.set(key, encape, ctag)
        return encape, ctag, 1
    def make_conf(self, conf, deal_ns=None, id=None):
        deal_key, dk_find = self.deal_key(conf)
        deal, dtag, dfind = self.deals.tget(deal_key, deal_ns, id, True)
        assert dfind, f'deal for {deal_ns}/{deal_key} not found'
        encape, conf, upd = deal(conf, self.unit)
        return encape, conf, upd
    def tget(self, key, src=None,id=None, gfind=True):
        if Confs.is_conf(key):
            encape, conf, upd = self.make_conf(key, self.deals.ns, id)
            return encape, Datas.Key.Pub, 1
        ns, id = self.nsid(src, id)
        obj, tag, find=super().tget(key, ns, id, False)
        if find:
            return obj, tag, find
        obj, tag, find = self.make(key, ns, id)
        if find:
            return obj, tag, find
        if not gfind:
            return 0,0,0
        return self.dts.tget(key, ns, id)
class Encapeset(Dataset):
    '''
        生成对象前的中间层
    '''
    def init(self, ids, mg=None):
        super().init(ids)
        self.confs = None
        self.deals = None
        self.deal_key = None
        self.mg = None
        self.bind(mg)
    def bind(self, mg):
        if self.mg==mg:
            return self
        self.mg = mg
        self.confs = mg.confs
        self.deals = mg.deals
        self.deal_key = mg.deal_key
        return self
    def make(self, key, ns=None, tag=None, id = None):
        conf, keys, ctag, cfind = self.confs.get(key, ns, tag, id)
        if not cfind:
            return None,None,None,0
        conf, deal_ns, cid = conf
        encape, conf, upd  = self.make_conf(conf, deal_ns, id, cid)
        if cid in self.objs:
            obj = self.objs[cid]
            if ns != obj.ns:
                ids_ns = self.ids(obj.ns)
                keys_ns = keys[:len(ids_ns)]
                assert ids_ns == keys_ns, f"{ids_ns}!={keys_ns}"
                keys = keys[len(ids_ns):]
                key = self.ids.id(keys)
            self.objs[cid].set(key, encape, ctag)
            if upd:
                self.confs.objs[cid].set(key, conf, ctag)
        else:
            self.set(key, encape, ns, ctag)
        return encape, keys, ctag, 1
    def make_conf(self, conf, deal_ns=None, id=None,cid=None):
        deal_key, dk_find = self.deal_key(conf)
        deal, dtag, dfind = self.deals.tget(deal_key, deal_ns, id)
        assert dfind, f"encapes make_conf deal {deal_ns}/'{deal_key}' not found"
        unit = self.mg.get_unit(cid)
        encape, conf, upd = deal(conf, unit)
        return encape, conf, upd 
    def get(self, key, ns=None, tag=None, id=None):
        if Confs.is_conf(key):
            encape, conf, upd = self.make_conf(key, None, id)
            return encape, None, None, Datas.Key.Pub, 1
        obj, keys, _tag, find = super().get(key, ns, tag, id)
        if find:
            return obj, keys, _tag, find
        return self.make(key, ns, tag, id)

    


from buildz import Base, xf
from . import argz
from . import build
from . import evalx
class Fc(Base):
    def str(self):
        return str(self.fc)
    def init(self, fc=None):
        self.fc = fc
    def call(self, args, maps):
        return self.fc(*args, **maps)
    @staticmethod
    def make(fc):
        if isinstance(fc, Fc):
            return fc
        return Fc(fc)
class RetFc(Fc):
    def call(self, args, maps):
        return args, maps
class ArgsCall(Fc):
    def init(self, fc, args=None, name = None):
        fc=self.make(fc)
        if name is None:
            name = str(fc)
        self.name = name
        self.fc = fc
        self.args = args
    def call(self, args, maps):
        if self.args is not None:
            args, maps = self.args(args, maps)
        try:
            return self.fc(args, maps)
        except argz.ArgExp as exp:
            if self.args is not None:
                exp = self.args.deal_exp(exp)
            raise exp
class Calls(Fc):
    def str(self):
        return "argz.Calls"
    def init(self, fcs = None, args=None, name = None):
        if name is None:
            name = "argz.calls"
        self.name = name
        if fcs is None:
            fcs = []
        self.fcs = fcs
        self.args = args
    def add(self, fc):
        fc=self.make(fc)
        self.fcs.append(fc)
    def call(self, args, maps):
        val = None
        if self.args is not None:
            args, maps = self.args(args, maps)
        try:
            for fc in self.fcs:
                val = fc(args, maps)
            return val
        except argz.ArgExp as exp:
            if self.args is not None:
                exp = self.args.deal_exp(exp)
            raise exp
class EvalCall(Fc):
    def init(self, fc, eval=None, name = None):
        fc=self.make(fc)
        if name is None:
            name = str(fc)
        self.eval = eval
        self.fc = fc
        self.name = name
    def call(self, args, maps):
        if self.eval is not None:
            if not self.eval([args, maps]):
                return None
        return self.fc(args, maps)
class CallsBuild(build.Build):
    def call(self, conf):
        ctype = xf.g(conf, type=None)
        if ctype != "calls":
            return None
        calls = xf.g(conf, calls = [])
        if len(calls)==0:
            return None
        calls = [self.builder.get_conf(conf) for conf in calls]
        return Calls(calls)
class FcBuild(build.Build):
    def call(self, conf):
        ctype = xf.g(conf, type=None)
        if ctype != "call":
            return None
        src = xf.g(conf, src = None)
        fc = self.builder.get_var(src)
        return Fc(fc)
class RetBuild(build.Build):
    def call(self, conf):
        ctype = xf.g(conf, type=None)
        if ctype != "ret":
            return None
        return RetFc()
class CallBuild(build.Build):
    def init(self):
        self.args = argz.ArrArgsBuild()
        self.eval = evalx.EvalBuild()
        self.fc = FcBuild()
        self.calls = CallsBuild()
        self.ret = RetBuild()
        super().init(self.args, self.eval, self.fc, self.calls)
    def call(self, conf):
        args = self.args(conf)
        judges = self.eval(conf)
        fc = self.fc(conf)
        calls = self.calls(conf)
        ret = self.ret(conf)
        bfc = fc is None
        bcls = calls is None
        bret = ret is None
        assert (bfc+bcls+bret)==2
        if fc is None:
            fc = calls
        if fc is None:
            fc = ret
        if args is not None:
            fc = ArgsCall(fc, args)
        if judges is not None:
            fc = EvalCall(fc, judges)
        return fc
class CallBuilder(build.Builder):
    def init(self, key_id = "id"):
        super().init(CallBuild(), key_id)
